"""
Module to test telemetry main class for pipeline telemetry module.
"""
from datetime import datetime

import pytest
from test_data import DEFAULT_TELEMETRY_PARAMS

from pipeline_telemetry.main import FAIL_COUNT, Telemetry, mongo_telemetry
from pipeline_telemetry.settings import exceptions, settings
from pipeline_telemetry.settings.settings import ProcessType
from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage
from pipeline_telemetry.storage.mongo import TelemetryMongoStorage

# pylint: disable=protected-access


def test_telemetry_exists():
    """ check that Telemetry class exists """
    assert Telemetry


def test_telemetry_instance_creation():
    """ check that Telemetry instance can be created using default_params """
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).process_name == \
        'load_weather_data'


def test_telemetry_instance_creation_raises_excption():
    """
    Check that Telemetry instance creation raises exception with non registered
    ProcessType.
    """
    process_type = ProcessType(
        process_type='not_registered',
        subtypes=['test'])
    telemetry_params = {
        'process_name': 'load_weather_data',
        'process_type': process_type
    }
    with pytest.raises(exceptions.ProcessTypeNotRegistered):
        Telemetry(**telemetry_params)


def test_telemetry_instance_has_telemetry_property():
    """ check that Telemetry instance has a telemetry property """
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry


def test_telemetry_instance_has_sub_process_types_property():
    """
    check that Telemetry instance has a sub_process_types property that return
    a dict with all sub_process_types
    """
    assert isinstance(
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_process_types, list)
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_process_types == \
        settings.DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES


def test_telemetry_instance_raises_exception_with_invalid_process_type():
    """
    check that Telemetry instance creation raises exception when invalid
    process_type is used
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {'process_type': 'invalid'}
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry(**telemetry_params)


def test_telemetry_instance_sets_start_time():
    """
    check that Telemetry instance has a start_date_time
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry
    assert isinstance(telemetry.get('start_date_time'), datetime)


def test_increase_new_sub_process_base_count_to_telemetry():
    """
    Check that a new sub process base count can be added to a Telemetry
    instance.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    assert 'RETRIEVE_RAW_DATA' in telemetry_inst.telemetry
    assert telemetry_inst.get('RETRIEVE_RAW_DATA')['base_counter'] == 1


def test_increase_non_exsiting_sub_process_raises_exception():
    """
    Check that a add a based count for a non existing sub process raises
    an excpetion
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.InvalidSubProcess):
        telemetry_inst.increase_sub_process_base_count('Non exsiting')


def test_increase_sub_process_base_count_to_telemetry():
    """
    Check that an existing sub process base count can be inreased.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    assert telemetry_inst.get('RETRIEVE_RAW_DATA')['base_counter'] == 1
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    assert telemetry_inst.get('RETRIEVE_RAW_DATA')['base_counter'] == 2


def test_increase_sub_process_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')


def test_can_not_close_closed_telemetry_instance():
    """
    check that a closed telemetry instance can not be closed again
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.save_and_close()


def test_increase_sub_process_fail_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_fail_count('RETRIEVE_RAW_DATA')


def test_sub_process_custom_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process custom count method raises an excpetion
    when telementry object has been closed
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_custom_count(
            sub_process='RETRIEVE_RAW_DATA',
            custom_counter='test_counter')


def test_custom_count_not_allowed_with_closed_telemetry():
    """
    check that a custom count method raises an excpetion
    when telementry object has been closed
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_custom_count(
            custom_counter='test_counter')


def test_add_sub_process_fail_count_to_telemetry():
    """
    check that a sub process fail count can be increased
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.increase_sub_process_fail_count('RETRIEVE_RAW_DATA')
    assert telemetry_inst.telemetry['RETRIEVE_RAW_DATA'][FAIL_COUNT] == 1


def test_add_sub_process_fail_count_raises_excpetion():
    """
    check that a sub process fail count raises an excpetion when sub_process
    has not yet been initialized by increase_sub_process_base_count(sub_process)
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded):
        telemetry_inst.increase_sub_process_fail_count('RETRIEVE_RAW_DATA')


def test_add_custom_counter_to_telemetry():
    """
    check that a custom counter can be added to the telemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_custom_count('network_error')
    assert telemetry_inst.telemetry['network_error'] == 1


def test_add_sub_process_custom_counter_to_telemetry():
    """
    Check that a custom counter can be added to a subprocess.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.increase_sub_process_custom_count(
        custom_counter='network_error',
        sub_process='RETRIEVE_RAW_DATA')
    assert telemetry_inst.get('RETRIEVE_RAW_DATA')['network_error'] == 1


def test_increase_existimg_sub_process_custom_counter_to_telemetry():
    """
    Check that a existing custom counter can be icnrease
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    for _ in [1, 2]:
        telemetry_inst.increase_sub_process_custom_count(
            custom_counter='network_error',
            sub_process='RETRIEVE_RAW_DATA')

    assert telemetry_inst.get('RETRIEVE_RAW_DATA')['network_error'] == 2


def test_increase_custom_counter_to_telemetry():
    """
    check that a custom counter once created can be increased with any value
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_custom_count('network_error')
    telemetry_inst.increase_custom_count('network_error', 2)
    assert telemetry_inst.telemetry['network_error'] == 3


def test_increase_sub_process_custom_count_raises_excpetion():
    """
    check that increase_sub_process_custom_count method raises an excpetion
    when sub_process has not yet been initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded):
        telemetry_inst.increase_sub_process_custom_count(
            sub_process='RETRIEVE_RAW_DATA',
            custom_counter='test_counter')


def test_increase_sub_process_custom_count_with_two():
    """
    check that increase_sub_process_custom_count method creates counter
    with initial value 2
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_inst.increase_sub_process_custom_count(
        sub_process='RETRIEVE_RAW_DATA',
        custom_counter='test_counter',
        increment=2)
    assert telemetry_inst.telemetry[
        'RETRIEVE_RAW_DATA']['test_counter'] == 2


def test_close_telemetry_instance_sets_run_time(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        ("pipeline_telemetry.storage.memory."
         "TelemetryInMemoryStorage.store_telemetry"),
        return_value=None)
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry_result = telemetry_inst.save_and_close()
    assert float(telemetry_result.get('run_time_in_seconds')) > 0


def test_close_telemetry_instance_calls_store_telemetry(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        ("pipeline_telemetry.storage.memory."
         "TelemetryInMemoryStorage.store_telemetry"),
        return_value=None)
    _store_telemetry_spy = mocker.spy(
        TelemetryInMemoryStorage, 'store_telemetry')
    Telemetry(**DEFAULT_TELEMETRY_PARAMS).save_and_close()
    assert _store_telemetry_spy.called


def test_get_storage_class_returns_in_memory_storage_by_default():
    """
    check that _get_storage_class method returns TelemetryInMemoryStorage
    instance if no storage class is provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert isinstance(
        telemetry_inst._get_storage_class(None), TelemetryInMemoryStorage)


def test_get_storage_class_raises_exception_if_invalid_class_is_provided():
    """
    check that _get_storage_class raises exception if storage class is provided
    that is not a subclass of AbstractTelemetryStorage
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.StorageClassOfIncorrectType):
        telemetry_inst._get_storage_class(str)


def test_get_storage_class_returns_instance_of_storage_class():
    """
    check that _get_storage_class method returns an instance of
    Storage class provided
    """
    # pylint: disable=too-few-public-methods
    class TestStorage(AbstractTelemetryStorage):
        """ test class """

        def store_telemetry(self, telemetry):
            """ test method """
            return None

    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert isinstance(
        telemetry_inst._get_storage_class(TestStorage), TestStorage)


def test_storage_class_close_method_closes_db():
    """
    check that _get_storage_class method returns an instance of
    Storage class provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    storage_instance = telemetry_inst._get_storage_class(None)
    storage_instance.close_db()
    assert storage_instance.db_in_memory is None
    assert storage_instance.db_cursor is None


def test_mongo_telemetry():
    """
    Test mongo_telemetry method returns a Telemetry object with a mongo storage
    class.
    """
    telemetry = mongo_telemetry(telemetry_rules={}, **DEFAULT_TELEMETRY_PARAMS)
    assert isinstance(telemetry, Telemetry)
    assert isinstance(telemetry._storage_class, TelemetryMongoStorage)



"""[summary]
"""
import json
import sqlite3

from .generic import AbstractTelemetryStorage


class TelemetryInMemoryStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in memory storage for use when unit testing
    """
    db_in_memory = None
    db_cursor = None

    def __init__(self):
        if not self.db_in_memory:
            self.initialize_db()

    @classmethod
    def initialize_db(cls):
        """
        class method to initialize the in memory db
        """
        if not cls.db_in_memory:
            cls.db_in_memory = \
                sqlite3.connect(":memory:")
            cls.db_cursor = cls.db_in_memory.cursor()
            cls._define_db_table(cls.db_cursor)

    @classmethod
    def close_db(cls):
        """close cursor and connections
        """
        cls.db_cursor.close()
        cls.db_in_memory.close()
        cls.db_cursor = None
        cls.db_in_memory = None

    @staticmethod
    def _define_db_table(cursor):
        """ define telemetry table """
        cursor.executescript('''
        DROP TABLE IF EXISTS telemetry;
        CREATE TABLE telemetry (process_name varchar(40),
        process_type varchar(40), start_date_time varchar(30),
        run_time varchar(20), telemetry_data json)''')

    def store_telemetry(self, telemetry: dict) -> None:
        """ public method to persist telemetry object"""
        telemetry_copy = telemetry.copy()
        process_name = telemetry_copy.pop('process_name', None)
        process_type = telemetry_copy.pop('process_type', None)
        start_date_time = telemetry_copy.pop('start_date_time', None)
        run_time_in_seconds = telemetry_copy.pop('run_time_in_seconds', None)
        json_object = json.dumps(telemetry_copy)

        self.db_cursor.execute(
            "insert into telemetry values (?, ?, ?, ?, ?)",
            [process_name, process_type, start_date_time,
             run_time_in_seconds, json_object]
        )



from dataclasses import dataclass


@dataclass(frozen=True)
class IntegrationTestData():
    """Immutable dataclass to define integration test data."""
    data: dict
    validation_rules: dict
    telemetry_result: dict

    @property
    def result_field(self):
        return self.telemetry_result.get('field')

    @property
    def expected_result(self):
        return self.telemetry_result.get('result')


def validate_result_from_telemetry(
        telemetry: dict, test_data: IntegrationTestData) -> bool:
    """
    Validates if the telemetry data is according to expected results as defined
    in IntegrationTestData object test_data
    """
    result_field = test_data.result_field
    expected_result = test_data.expected_result
    return telemetry.get(result_field) == expected_result


BASIC_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={'items': [
        {'id': 1, 'field1': 'test'},
        {'id': 2, 'field1': 'test'},
        {'id': 3, 'field1': 'test'},
        {'id': 4, 'field1': 'test'}]},
    validation_rules={
        'RETRIEVE_RAW_DATA': {
            'has_key': {'field_name': 'items'},
            'validate_entries': {
                'field_name': 'items',
                'expected_count': 4}}},
    telemetry_result={
        'field': 'RETRIEVE_RAW_DATA',
        'result': {'base_counter': 1, 'fail_counter': 0}})

MISSING_KEY_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={'no items key': []},
    validation_rules={
        'RETRIEVE_RAW_DATA': {
            'has_key': {'field_name': 'items'}}},
    telemetry_result={
        'field': 'RETRIEVE_RAW_DATA',
        'result': {
            'base_counter': 1,
            'fail_counter': 0,
            'HAS_KEY_ERR_0001@KEY_<items>': 1}})


MUST_HAVE_KEY_IN_ITEMS_TEST = IntegrationTestData(
    data={'items': [{'key': 1}, {'no_key': 2}]},
    validation_rules={
        'RETRIEVE_RAW_DATA': {
            'entries_have_key': {'field_name': 'items', 'must_have_key': 'key'}}},
    telemetry_result={
        'field': 'RETRIEVE_RAW_DATA',
        'result': {
            'base_counter': 1,
            'fail_counter': 0,
            'ENTRIES_HAVE_KEY_ERR_003@KEY_<items_key>': 1}})

NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST = IntegrationTestData(
    data={'items': [{'no_key': 1}, {'no_key': 2}]},
    validation_rules={
        'RETRIEVE_RAW_DATA': {
            'entries_have_key': {'field_name': 'items', 'must_have_key': 'key'}}},
    telemetry_result={
        'field': 'RETRIEVE_RAW_DATA',
        'result': {
            'base_counter': 1,
            'fail_counter': 0,
            'ENTRIES_HAVE_KEY_ERR_003@KEY_<items_key>': 2}})

ITEMS_FIELD_HAS_DICT_TEST = IntegrationTestData(
    data={'items': {'key': 1, 'no_key': 2}},
    validation_rules={
        'RETRIEVE_RAW_DATA': {
            'entries_have_key': {'field_name': 'items', 'must_have_key': 'key'}}},
    telemetry_result={
        'field': 'RETRIEVE_RAW_DATA',
        'result': {
            'base_counter': 1,
            'fail_counter': 0,
            'ENTRIES_HAVE_KEY_ERR_002@KEY_<items>': 1}})



"""
Module to define customer excpetions for data_validator module

exceptions
- FieldNameMandatory
- UnknownInstruction
- InstructionRegisteredTwice
- RuleCanHaveOnlyOneInstruction
- BaseCountForSubProcessNotAdded
- TelemetryObjectAlreadyClosed
- StorageClassOfIncorrectType
- ProcessTypeMustBeOfClassProcessType
- ProcessTypeNotRegistered
"""


class FieldNameMandatory(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, instruction):
        message = f'For `{instruction}` instruction'
        super().__init__(message)


class MustHaveKeyMandatory(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, instruction):
        message = f'For `{instruction}` instruction'
        super().__init__(message)


class ExpectedCountMustBePositiveInt(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self):
        message = ''.join([
            'Ruleset does not contain `expected_count` or ',
            'expected count is negative'])
        super().__init__(message)


class UnknownInstruction(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, instruction):
        super().__init__(f'{instruction} not registered')


class InstructionRegisteredTwice(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, instruction_class):
        message = ''.join([
            f'{instruction_class.instruction} from class ',
            f'{instruction_class.__name__}.'
        ])
        super().__init__(message)


class RuleCanHaveOnlyOneInstruction(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, rule):
        message = ''.join([
            'Rule contains multiple keys / instruction : ',
            ', '.join(rule.keys())
        ])
        super().__init__(message)


class InvalidSubProcess(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, sub_process: str, process_type):
        message = ''.join([
            f'Sub Process `{sub_process}` does not exist for ',
            f'process type `{process_type.name}`'])
        super().__init__(message)


class BaseCountForSubProcessNotAdded(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, sub_process):
        message = ''.join([
            f'Sub process {sub_process} has not yet been initialized.',
            f' First call increase_sub_process_base_count({sub_process}) ',
            'on telemetry instance'])
        super().__init__(message)


class TelemetryObjectAlreadyClosed(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self):
        message = 'Telemetry object is already closed'
        super().__init__(message)


class StorageClassOfIncorrectType(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, class_name):
        message = ''.join([
            f'StorageClass `{class_name}` not a child ',
            'class of AbstractTelemetryStorage'])
        super().__init__(message)


class ProcessTypeMustBeOfClassProcessType(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self):
        message = 'provide process_type not of class ProcessType'
        super().__init__(message)


class ProcessTypeNotRegistered(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self, process_type):
        message = f'provided process_type {process_type.name} not registered'
        super().__init__(message)


class ProcessTypesMustBeOfClassBaseEnumertor(Exception):
    """ custom exception for Telemetry Module """

    def __init__(self):
        message = 'provided process_types enumerator not of class BaseEnumrator'
        super().__init__(message)



"""
This module provides test data for the telemetry tests
"""
from errors.base import ErrorCode

from pipeline_telemetry.settings.process_type import ProcessTypes
from pipeline_telemetry.settings.settings import \
    DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES, ProcessType


class InstructionTestClass():
    """Instruction class for test purposes."""
    # pylint: disable=too-few-public-methods
    instruction = 'test_instruction'
    fieldname = 'field_name'

    def __new__(cls):
        """ make this a singleton class """
        return cls


# default params for creating a test Telemetry object
DEFAULT_TELEMETRY_PARAMS = {
    'process_name': 'load_weather_data',
    'process_type': ProcessTypes.CREATE_DATA_FROM_URL
}

# custom process type for testing extending the default process types
CUSTOM_PROCESS_TYPE = {
    'custom_process_type': ['CUSTOM_SUB_TYPE']}

TEST_ERROR_CODE = ErrorCode(
    code="TEST_CODE_0001",
    description='test error'
)

TEST_TELEMETRY_RULES = {
    'RETRIEVE_RAW_DATA': {
        'has_key': {'field_name': 'items'}
    }
}

TEST_PROCESS_TYPE = ProcessType(
    process_type='test_process_type',
    subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES)



"""[summary]
"""
from dataclasses import dataclass
from enum import Enum
from typing import List

DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES = [
    "RETRIEVE_RAW_DATA",
    "DATA_CONVERSION",
    "DATA_STORAGE",
]

DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES = [
    "DATA_SELECTION",
    "DATA_CONVERSION",
    "DATA_UPLOAD",
]


@dataclass(frozen=True)
class ProcessType:
    """Immutable dataclass to define a process type and its subtypes."""

    process_type: str
    subtypes: list = list

    @property
    def name(self) -> str:
        """Property method to return process key name."""
        return self.process_type

    @property
    def sub_processes(self) -> List[str]:
        """Property method to return list of sub process types."""
        return self.subtypes


class BaseEnumerator(Enum):
    """
    Class to define base enumerator with a keys method.
    Only to be used for defining new enumerators.
    """

    @classmethod
    def keys(cls):
        return list(cls.__members__.keys())


class DefaultProcessTypes(BaseEnumerator):
    """
    Class to define the defailt process types with their subtypes
    """

    CREATE_DATA_FROM_URL = ProcessType(
        process_type="create_data_from_url",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_API = ProcessType(
        process_type="create_data_from_api",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_FILE = ProcessType(
        process_type="create_data_from_file",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    UPLOAD_DATA = ProcessType(
        process_type="upload_data", subtypes=DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES
    )



"""Module to test storage module.
"""
import sqlite3

from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage
from pipeline_telemetry.storage.mongo import TelemetryMongoModel, \
    TelemetryMongoStorage
from pipeline_telemetry.storage.mongo_connection import get_mongo_db_port


def test_abstract_strorage_class_exists():
    """Test abstract storage class exists."""
    assert AbstractTelemetryStorage


def test_in_memory_strorage_class_exists():
    """Test that in memory storage class exists."""
    assert TelemetryInMemoryStorage


def test_in_memory_strorage_creates_db_in_memory():
    """Test in memory storage instance has db_in_memory and db_cursor """
    in_memory_storage = TelemetryInMemoryStorage()
    assert in_memory_storage.db_in_memory
    assert in_memory_storage.db_cursor
    assert isinstance(in_memory_storage.db_in_memory, sqlite3.Connection)
    assert isinstance(in_memory_storage.db_cursor, sqlite3.Cursor)


def test_store_telemetry_stores_object():
    """Test in memory storage instance has db_in_memory and db_cursor """
    telemetry = {'process_name': 'test'}
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(
        TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(telemetry)
    all_in_memory_objects = \
        in_memory_storage.db_cursor.execute('SELECT * FROM telemetry ')
    assert len(all_in_memory_objects.fetchall()) == 1


def test_telemetry_mongo_model_class_exists():
    """Test TelemetryMongoModel class exists."""
    assert TelemetryMongoModel


def test_telemetry_mongo_storage_class_exists():
    """Test TelemetryMongoStorage class exists."""
    assert TelemetryMongoStorage


def test_store_telemetry_method_processes_and_saves_telemetry(mocker):
    """
    Test store_telemetry method calls _telemetry_model_kwargs and then creates
    and saves a mongo object.
    """
    mongo_module_path = "pipeline_telemetry.storage.mongo."
    mocker.patch(
        mongo_module_path + "TelemetryMongoModel.save", return_value=True)
    mocker.patch(
        mongo_module_path + "TelemetryMongoStorage._telemetry_model_kwargs",
        return_value={})
    _telemetry_model_kwargs_spy = mocker.spy(
        TelemetryMongoStorage, '_telemetry_model_kwargs')
    _telemetry_mongo_model_new_spy = mocker.spy(
        TelemetryMongoModel, '__new__')
    _telemetry_mongo_model_save_spy = mocker.spy(
        TelemetryMongoModel, 'save')
    TelemetryMongoStorage().store_telemetry(
        telemetry={'process_name': 'test'})
    assert _telemetry_model_kwargs_spy.called
    assert _telemetry_mongo_model_new_spy.called
    assert _telemetry_mongo_model_save_spy.called


def test_telemetry_model_kwargs_method():
    """Test _telemetry_model_kwargs method returns correct kwargs."""
    result = TelemetryMongoStorage()._telemetry_model_kwargs({
        'process_name': 'tst_process_name',
        'process_type': 'tst_process_type',
        'start_date_time': 'tst_start_date_time',
        'run_time_in_seconds': 'tst_run_time_in_seconds',
        'field1': {'a': 1},
        'field2': 'value'})

    assert result == {
        'process_name': 'tst_process_name',
        'process_type': 'tst_process_type',
        'start_date_time': 'tst_start_date_time',
        'run_time_in_seconds': 'tst_run_time_in_seconds',
        'telemetry': {
            'field1': {'a': 1},
            'field2': 'value'}}


def test_get_mongo_db_port_returns_none():
    """Test _get_mongo_db_port returns None if no env variable is set."""
    assert get_mongo_db_port() is None


def test_get_mongo_db_port_returns_port_int(mocker):
    """Test _get_mongo_db_port returns port nr as intno env variable is set."""
    mocker.patch("os.getenv", return_value='12345')
    assert get_mongo_db_port() == 12345


def test_default_access_params():
    """Test default_access_params are set poperly."""
    from pipeline_telemetry.storage import mongo_connection as mc
    def_params = mc.MONGO_ACCESS_PARAMS
    assert def_params['port'] == mc.DEFAULT_MONGO_DB_PORT
    assert def_params['host'] == mc.DEFAULT_MONGO_DB_HOST



"""Module to provide a storage class for using MongoDB.

When using MongoDB storage class you can select the mongoDB instance to be used
via environment Variables

MONGO_DB_NAME (defaults to `telemetry`)
MONGO_DB_PASSWORD
MONGO_DB_USERNAME
MONGO_DB_HOST
MONGO_DB_PORT

If no host and port are defined the connection will dedault to a localhost
mongoDB instance.
"""
from mongoengine import DictField, Document, StringField, connect

from .generic import AbstractTelemetryStorage
from .mongo_connection import MONGO_ACCESS_PARAMS

connect(**MONGO_ACCESS_PARAMS)


class TelemetryMongoModel(Document):
    """
    Class to provice telemetry Mongo Model for persistance in MongoDB
    """
    process_name = StringField()
    process_type = StringField()
    start_date_time = StringField()
    run_time_in_seconds = StringField()
    telemetry = DictField(default=None)


class TelemetryMongoStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in mongo storage class.
    This class can be used as storage_class argument when creating
    an instance of Telemetry.
    """

    def store_telemetry(self, telemetry: dict) -> None:
        """ public method to persist telemetry object"""
        telemetry_mongo_kwargs = self._telemetry_model_kwargs(telemetry)
        TelemetryMongoModel(**telemetry_mongo_kwargs).save()

    @staticmethod
    def _telemetry_model_kwargs(telemetry: dict) -> dict:
        """
        Returns a dicts with kwargs that can be used to create a new
        TelemetryMongoStorage instance.
        """
        telemetry_copy = telemetry.copy()
        process_name = telemetry_copy.pop('process_name', None)
        process_type = telemetry_copy.pop('process_type', None)
        start_date_time = telemetry_copy.pop('start_date_time', None)
        run_time_in_seconds = telemetry_copy.pop('run_time_in_seconds', None)
        return {
            'process_name': process_name,
            'process_type': process_type,
            'start_date_time': start_date_time,
            'run_time_in_seconds': run_time_in_seconds,
            'telemetry': telemetry_copy
        }
